@@ -25,7 +25,7 @@ This module encapsulates an email message and allows access to the various MIME
 message parts and message metadata.
 
 The message structure, after initiating a parse() cycle, looks like this:
- 
+
   Message object, also top-level node in Message::Node tree
      |
      +---> Message::Node for other parts in MIME structure
@@ -44,6 +44,12 @@ package Mail::SpamAssassin::Message;
 
 use strict;
 use warnings;
+use re 'taint';
+
+BEGIN {
+  eval { require Digest::SHA; import Digest::SHA qw(sha1 sha1_hex); 1 }
+  or do { require Digest::SHA1; import Digest::SHA1 qw(sha1 sha1_hex) }
+}
 
 use Mail::SpamAssassin;
 use Mail::SpamAssassin::Message::Node;
@@ -62,10 +68,10 @@ use vars qw(@ISA);
 Creates a Mail::SpamAssassin::Message object.  Takes a hash reference
 as a parameter.  The used hash key/value pairs are as follows:
 
-C<message> is either undef (which will use STDIN), a scalar of the
-entire message, an array reference of the message with 1 line per array
-element, and either a file glob or IO::File object which holds the entire
-contents of the message.
+C<message> is either undef (which will use STDIN), a scalar - a string
+containing an entire message, a reference to such string, an array reference
+of the message with one line per array element, or either a file glob
+or an IO::File object which holds the entire contents of the message.
 
 Note: The message is expected to generally be in RFC 2822 format, optionally
 including an mbox message separator line (the "From " line) as the first line.
@@ -97,7 +103,7 @@ sub new {
   $class = ref($class) || $class;
 
   my($opts) = @_;
-  my $message = $opts->{'message'} || \*STDIN;
+  my $message = defined $opts->{'message'} ? $opts->{'message'} : \*STDIN;
   my $parsenow = $opts->{'parsenow'} || 0;
   my $normalize = $opts->{'normalize'} || 0;
 
@@ -113,6 +119,27 @@ sub new {
   $self->{pristine_body} =	'';
   $self->{mime_boundary_state} = {};
   $self->{line_ending} =	"\012";
+  $self->{master_deadline} = $opts->{'master_deadline'};
+  $self->{suppl_attrib} = $opts->{'suppl_attrib'};
+
+  if ($self->{suppl_attrib}) {  # caller-provided additional information
+    # pristine_body_length is currently used by an eval test check_body_length
+    # Possible To-Do: Base the length on the @message array later down?
+    if (defined $self->{suppl_attrib}{body_size}) {
+      # Optional info provided by a caller; should reflect the original
+      # message body size if provided, and as such it may differ from the
+      # $self->{pristine_body} size, e.g. when the caller passed a truncated
+      # message to SpamAssassin, or when counting line-endings differently.
+      $self->{pristine_body_length} = $self->{suppl_attrib}{body_size};
+    }
+    if (ref $self->{suppl_attrib}{mimepart_digests}) {
+      # Optional info provided by a caller: an array of digest codes (e.g. SHA1)
+      # of each MIME part. Should reflect the original message if provided.
+      # As such it may differ from digests calculated by get_mimepart_digests(),
+      # e.g. when the caller passed a truncated message to SpamAssassin.
+      $self->{mimepart_digests} = $self->{suppl_attrib}{mimepart_digests};
+    }
+  }
 
   bless($self,$class);
 
@@ -129,16 +156,38 @@ sub new {
   if (ref $message eq 'ARRAY') {
      @message = @{$message};
   }
-  elsif (ref $message eq 'GLOB' || ref $message eq 'IO::File') {
+  elsif (ref($message) eq 'GLOB' || ref($message) =~ /^IO::/) {
     if (defined fileno $message) {
-      @message = <$message>;
+
+      # sysread+split avoids a Perl I/O bug (Bug 5985)
+      # and is faster than (<$message>) by 10..25 %
+      # (a drawback is a short-term double storage of a text in $raw_str)
+      #
+      my($nread,$raw_str); $raw_str = '';
+      while ( $nread=sysread($message, $raw_str, 16384, length $raw_str) ) { }
+      defined $nread  or die "error reading: $!";
+      @message = split(/^/m, $raw_str, -1);
+
+      if ($raw_str eq '') {
+        dbg("message: empty message read");
+      } elsif (length($raw_str) > 128*1024) {
+        # ditch rarely used large chunks of allocated memory, Bug 6514
+        #   http://www.perlmonks.org/?node_id=803515
+        # about 97% of mail messages are below 128 kB,
+        # about 98% of mail messages are below 256 kB (2010 statistics)
+        # dbg("message: deallocating %.2f MB", length($raw_str)/1024/1024);
+        undef $raw_str;
+      }
     }
   }
+  elsif (ref $message eq 'SCALAR') {
+    @message = split(/^/m, $$message, -1);
+  }
   elsif (ref $message) {
     dbg("message: Input is a reference of unknown type!");
   }
   elsif (defined $message) {
-    @message = split ( /^/m, $message );
+    @message = split(/^/m, $message, -1);
   }
 
   # Pull off mbox and mbx separators
@@ -148,7 +197,8 @@ sub new {
     # if we get here, it means that the input was null, so fake the message
     # content as a single newline...
     @message = ("\n");
-  } elsif ($message[0] =~ /^From\s/) {
+  } elsif ($message[0] =~ /^From\s+(?!:)/) {
+    # careful not to confuse with obsolete syntax which allowed WSP before ':'
     # mbox formated mailbox
     $self->{'mbox_sep'} = shift @message;
   } elsif ($message[0] =~ MBX_SEPARATOR) {
@@ -158,7 +208,7 @@ sub new {
     # de facto portability standard in SA's internals.  We need to
     # to this so that Mail::SpamAssassin::Util::parse_rfc822_date
     # can parse the date string...
-    if (/([\s|\d]\d)-([a-zA-Z]{3})-(\d{4})\s(\d{2}):(\d{2}):(\d{2})/) {
+    if (/([\s\d]\d)-([a-zA-Z]{3})-(\d{4})\s(\d{2}):(\d{2}):(\d{2})/) {
       # $1 = day of month
       # $2 = month (text)
       # $3 = year
@@ -168,10 +218,10 @@ sub new {
       my @arr = localtime(timelocal($6,$5,$4,$1,$MONTH{lc($2)}-1,$3));
       my $address;
       foreach (@message) {
-  	if (/From:\s[^<]+<([^>]+)>/) {
+  	if (/^From:[^<]*<([^>]+)>/) {
   	    $address = $1;
   	    last;
-  	} elsif (/From:\s([^<^>]+)/) {
+  	} elsif (/^From:\s*([^<> ]+)/) {
   	    $address = $1;
   	    last;
   	}
@@ -182,20 +232,23 @@ sub new {
 
   # bug 4363
   # Check to see if we should do CRLF instead of just LF
-  # For now, just check the first header and do whatever it does
-  if (@message && $message[0] =~ /\015\012/) {
+  # For now, just check the first and last line and do whatever it does
+  if (@message && ($message[0] =~ /\015\012/ || $message[-1] =~ /\015\012/)) {
     $self->{line_ending} = "\015\012";
     dbg("message: line ending changed to CRLF");
   }
 
-  # Go through all the headers of the message
-  my $header = '';
-  while ( my $current = shift @message ) {
-    unless ($self->{'missing_head_body_separator'}) {
-      $self->{'pristine_headers'} .= $current;
-    }
+  # Is a CRLF -> LF line endings conversion necessary?
+  my $squash_crlf = $self->{line_ending} eq "\015\012";
+
+  # Go through all the header fields of the message
+  my $hdr_errors = 0;
+  my $header;
+  for (;;) {
+    # make sure not to lose the last header field when there is no body
+    my $eof = !@message;
+    my $current = $eof ? "\n" : shift @message;
 
-    # NB: Really need to figure out special folding rules here!
     if ( $current =~ /^[ \t]/ ) {
       # This wasn't useful in terms of a rule, but we may want to treat it
       # specially at some point.  Perhaps ignore it?
@@ -203,18 +256,20 @@ sub new {
       #  $self->{'obsolete_folding_whitespace'} = 1;
       #}
 
-      # append continuations if there's a header in process
-      if ($header) {
-        $header .= $current;
-      }
+      $header = ''  if !defined $header;  # header starts with a continuation!?
+      $header .= $current;  # append continuations, no matter what
+      $self->{'pristine_headers'} .= $current;
     }
-    else {
+    else {  # not a continuation
       # Ok, there's a header here, let's go ahead and add it in.
-      if ($header) {
+      if (defined $header) {  # deal with a previous header field
         my ($key, $value) = split (/:/s, $header, 2);
 
-        # If it's not a valid header (aka: not in the form "foo: bar"), skip it.
+        # If it's not a valid header (aka: not in the form "foo:bar"), skip it.
         if (defined $value) {
+	  # CRLF -> LF line-endings conversion if necessary
+	  $value =~ s/\015\012/\012/sg  if $squash_crlf;
+	  $key =~ s/[ \t]+\z//;  # strip WSP before colon, obsolete rfc822 syn
 	  # limit the length of the pairs we store
 	  if (length($key) > MAX_HEADER_KEY_LENGTH) {
 	    $key = substr($key, 0, MAX_HEADER_KEY_LENGTH);
@@ -228,31 +283,40 @@ sub new {
         }
       }
 
-      # not a continuation...
-      $header = $current;
-    }
-
-    if ($header) {
-      if ($header =~ /^\r?$/) {
-        last;
+      if ($current =~ /^\r?$/) {  # a regular end of a header section
+	if ($eof) {
+	  $self->{'missing_head_body_separator'} = 1;
+	} else {
+	  $self->{'pristine_headers'} .= $current;
+	}
+	last;
       }
+      elsif ($current =~ /^--/) {  # mime boundary encountered, bail out
+	$self->{'missing_head_body_separator'} = 1;
+	unshift(@message, $current);
+ 	last;
+      }
+      # should we assume entering a body on encountering invalid header field?
       else {
-        # Check for missing head/body separator
-	# RFC 2822, s2.2:
-	# A field name MUST be composed of printable US-ASCII characters
-	# (i.e., characters that have values between 33 (041) and 126 (176), inclusive),
-	# except colon (072).
-	# FOR THIS NEXT PART: list off the valid REs for what can be next:
-	#	Header, header continuation, blank line
-        if (!@message || $message[0] !~ /^(?:[\041-\071\073-\176]+:|[ \t]|\r?$)/ || $message[0] =~ /^--/) {
-	  # No body or no separator before mime boundary is invalid
-          $self->{'missing_head_body_separator'} = 1;
-	  
-	  # we *have* to go back through again to make sure we catch the last
-	  # header, so fake a separator and loop again.
-	  unshift(@message, "\n");
-        }
+        # no re "strict";  # since perl 5.21.8: Ranges of ASCII printables...
+        if ($current !~ /^[\041-\071\073-\176]+[ \t]*:/) {
+	  # A field name MUST be composed of printable US-ASCII characters
+	  # (i.e., characters that have values between 33 (041) and 126 (176),
+	  # inclusive), except colon (072). Obsolete header field syntax
+	  # allowed WSP before a colon.
+	  if (++$hdr_errors <= 3) {
+	    # just consume but ignore a few invalid header fields
+	  } else {  # enough is enough...
+	    $self->{'missing_head_body_separator'} = 1;
+	    unshift(@message, $current);
+ 	    last;
+	  }
+	}
       }
+
+      # start collecting a new header field
+      $header = $current;
+      $self->{'pristine_headers'} .= $current;
     }
   }
   undef $header;
@@ -261,15 +325,20 @@ sub new {
   # will get modified below
   $self->{'pristine_body'} = join('', @message);
 
-  # CRLF -> LF
-  # also merge multiple blank lines into a single one
-  my $start;
+  if (!defined $self->{pristine_body_length}) {
+    $self->{'pristine_body_length'} = length $self->{'pristine_body'};
+  }
+
   # iterate over lines in reverse order
+  # merge multiple blank lines into a single one
+  my $start;
   for (my $cnt=$#message; $cnt>=0; $cnt--) {
-    $message[$cnt] =~ s/\015\012/\012/;
+    # CRLF -> LF line-endings conversion if necessary
+    $message[$cnt] =~ s/\015\012\z/\012/  if $squash_crlf;
 
     # line is blank
-    if ($message[$cnt] !~ /\S/) {
+    if ($message[$cnt] =~ /^\s*$/) {
+      # /^\s*$/ is about 5% faster then !/\S/, but still expensive here
       if (!defined $start) {
         $start=$cnt;
       }
@@ -280,9 +349,10 @@ sub new {
 
     # if we've got a series of blank lines, get rid of them
     if (defined $start) {
+      my $max_blank_lines = 20;
       my $num = $start-$cnt;
-      if ($num > 10) {
-        splice @message, $cnt+2, $num-1;
+      if ($num > $max_blank_lines) {
+        splice @message, $cnt+2, $num-$max_blank_lines;
       }
       undef $start;
     }
@@ -293,6 +363,24 @@ sub new {
   ($self->{'type'}, $boundary) = Mail::SpamAssassin::Util::parse_content_type($self->header('content-type'));
   dbg("message: main message type: ".$self->{'type'});
 
+#  dbg("message: \$message[0]: \"" . $message[0] . "\"");
+
+  # bug 6845: if main message type is multipart and the message body does not begin with
+  # either a blank line or the boundary (if defined), insert a blank line
+  # to ensure proper parsing - do not consider MIME headers at the beginning of the body
+  # to be part of the message headers.
+  if ($self->{'type'} =~ /^multipart\//i && $#message > 0 && $message[0] =~ /\S/)
+  {
+    if (!defined $boundary || $message[0] !~ /^--\Q$boundary\E/)
+    {
+      dbg("message: Inserting blank line at top of body to ensure correct multipart MIME parsing");
+      unshift(@message, "\012");
+    }
+  }
+
+#  dbg("message: \$message[0]: \"" . $message[0] . "\"");
+#  dbg("message: \$message[1]: \"" . $message[1] . "\"");
+
   # parse queue, simple array of parts to parse:
   # 0: part object, already in the tree
   # 1: boundary used to focus body parsing
@@ -351,6 +439,10 @@ ie: If 'Subject' is specified as the header, and there are 2 Subject
 headers in a message, the last/bottom one in the message is returned in
 scalar context or both are returned in array context.
 
+Btw, returning the last header field (not the first) happens to be consistent
+with DKIM signatures, which search for and cover multiple header fields
+bottom-up according to the 'h' tag. Let's keep it this way.
+
 Note: the returned header will include the ending newline and any embedded
 whitespace folding.
 
@@ -359,20 +451,16 @@ whitespace folding.
 sub get_pristine_header {
   my ($self, $hdr) = @_;
   
-  return $self->{pristine_headers} unless $hdr;
-  my(@ret) = $self->{pristine_headers} =~ /^\Q$hdr\E:[ \t]+(.*?\n(?![ \t]))/smgi;
-  if (@ret) {
-    # ensure the response retains taintedness (bug 5283)
-    if (wantarray) {
-      return map {
-                Mail::SpamAssassin::Util::taint_var($_);
-              } @ret;
-    } else {
-      return Mail::SpamAssassin::Util::taint_var($ret[-1]);
-    }
-  }
-  else {
+  return $self->{pristine_headers} if !defined $hdr || $hdr eq '';
+  my(@ret) =
+    $self->{pristine_headers} =~ /^\Q$hdr\E[ \t]*:[ \t]*(.*?\n(?![ \t]))/smgi;
+  # taintedness is retained by "use re 'taint'" (fix in bug 5283 now redundant)
+  if (!@ret) {
     return $self->get_header($hdr);
+  } elsif (wantarray) {
+    return @ret;
+  } else {
+    return $ret[-1];
   }
 }
 
@@ -433,7 +521,7 @@ sub extract_message_metadata {
   my ($self, $permsgstatus) = @_;
 
   # do this only once per message, it can be expensive
-  if ($self->{already_extracted_metadata}) { return; }
+  return  if $self->{already_extracted_metadata};
   $self->{already_extracted_metadata} = 1;
 
   $self->{metadata}->extract ($self, $permsgstatus);
@@ -450,7 +538,10 @@ sub get_metadata {
   if (!$self->{metadata}) {
     warn "metadata: oops! get_metadata() called after finish_metadata()"; return;
   }
-  $self->{metadata}->{strings}->{$hdr};
+# dbg("message: get_metadata - %s: %s", $hdr, defined $_ ? $_ : '<undef>')
+#   for $self->{metadata}->{strings}->{lc $hdr};
+
+  $self->{metadata}->{strings}->{lc $hdr};
 }
 
 =item put_metadata($hdr, $text)
@@ -462,7 +553,8 @@ sub put_metadata {
   if (!$self->{metadata}) {
     warn "metadata: oops! put_metadata() called after finish_metadata()"; return;
   }
-  $self->{metadata}->{strings}->{$hdr} = $text;
+# dbg("message: put_metadata - %s: %s", $hdr, $text);
+  $self->{metadata}->{strings}->{lc $hdr} = $text;
 }
 
 =item delete_metadata($hdr)
@@ -474,7 +566,7 @@ sub delete_metadata {
   if (!$self->{metadata}) {
     warn "metadata: oops! delete_metadata() called after finish_metadata()"; return;
   }
-  delete $self->{metadata}->{strings}->{$hdr};
+  delete $self->{metadata}->{strings}->{lc $hdr};
 }
 
 =item $str = get_all_metadata()
@@ -487,9 +579,12 @@ sub get_all_metadata {
   if (!$self->{metadata}) {
     warn "metadata: oops! get_all_metadata() called after finish_metadata()"; return;
   }
-  my @ret = ();
-  foreach my $key (sort keys %{$self->{metadata}->{strings}}) {
-    push (@ret, "$key: " . $self->{metadata}->{strings}->{$key} . "\n");
+  my @ret;
+  my $keys_ref = $self->{metadata}->{strings};
+  foreach my $key (sort keys %$keys_ref) {
+    my $val = $keys_ref->{$key};
+    $val = ''  if !defined $val;
+    push (@ret, "$key: $val\n");
   }
   return (wantarray ? @ret :  join('', @ret));
 }
@@ -533,13 +628,22 @@ sub finish {
   delete $self->{'line_ending'};
   delete $self->{'missing_head_body_separator'};
 
+  # Remove the queue variable, in case the body has not been parsed
+  delete $self->{'parse_queue'};
+
   my @toclean = ( $self );
 
   # Go ahead and clean up all of the Message::Node parts
   while (my $part = shift @toclean) {
     # bug 5557: windows requires tmp file be closed before it can be rm'd
     if (ref $part->{'raw'} eq 'GLOB') {
-      close ($part->{'raw'});
+      close($part->{'raw'})  or die "error closing input file: $!";
+    }
+
+    # bug 5858: avoid memory leak with deep MIME structure
+    if (defined ($part->{metadata})) {
+      $part->{metadata}->finish();
+      delete $part->{metadata};
     }
 
     delete $part->{'headers'};
@@ -562,7 +666,9 @@ sub finish {
 
   # delete temporary files
   if ($self->{'tmpfiles'}) {
-    unlink @{$self->{'tmpfiles'}};
+    for my $fn (@{$self->{'tmpfiles'}}) {
+      unlink($fn) or warn "cannot unlink $fn: $!";
+    }
     delete $self->{'tmpfiles'};
   }
 }
@@ -571,8 +677,13 @@ sub finish {
 # temporary files are deleted even if the finish() method is omitted
 sub DESTROY {
   my $self = shift;
+  # best practices: prevent potential calls to eval and to system routines
+  # in code of a DESTROY method from clobbering global variables $@ and $! 
+  local($@,$!);  # keep outer error handling unaffected by DESTROY
   if ($self->{'tmpfiles'}) {
-    unlink @{$self->{'tmpfiles'}};
+    for my $fn (@{$self->{'tmpfiles'}}) {
+      unlink($fn) or dbg("message: cannot unlink $fn: $!");
+    }
   }
 }
 
@@ -632,34 +743,50 @@ sub parse_body {
     # one doesn't, assume it's malformed and send it to be parsed as a
     # non-multipart section
     #
-    if ( $toparse->[0]->{'type'} =~ /^multipart\//i && defined $toparse->[1] && ($toparse->[3] > 0)) {
+    my ($msg, $boundary, $body, $subparse) = @$toparse;
+
+    if ($msg->{'type'} =~ m{^multipart/}i && defined $boundary && $subparse > 0) {
       $self->_parse_multipart($toparse);
     }
     else {
       # If it's not multipart, go ahead and just deal with it.
       $self->_parse_normal($toparse);
 
-      if ($toparse->[0]->{'type'} =~ /^message\b/i && ($toparse->[3] > 0)) {
-        # Just decode the part, but we don't care about the result here.
-        $toparse->[0]->decode(0);
+      # bug 5041: process message/*, but exclude message/partial content types
+      if ($msg->{'type'} =~ m{^message/(?!partial\z)}i && $subparse > 0)
+      {
+        # Just decode the part, but we don't need the resulting string here.
+        $msg->decode(0);
 
-	# bug 5051: sometimes message/* parts have no content, and we get
-	# stuck waiting for STDIN, which is bad. :(
-        if ($toparse->[0]->{'decoded'}) {
-	  # Ok, so this part is still semi-recursive, since M::SA::Message calls
-	  # M::SA::Message, but we don't subparse the new message, and pull a
-	  # sneaky "steal our child's queue" maneuver to deal with it on our own
-	  # time.  Reference the decoded array directly since it's faster.
+        # bug 7125: decode and parse only message/rfc822 or message/global,
+        # but do not treat other message/* content types (like the ones listed
+        # here) as a message consisting of a header and a body, as they are not:
+        #    message/delivery-status, message/global-delivery-status,
+        #    message/feedback-report, message/global-headers,
+        #    message/global-disposition-notification,
+        #    message/disposition-notification, (and message/partial)
+
+        # bug 5051, bug 3748: check $msg->{decoded}: sometimes message/* parts
+        # have no content, and we get stuck waiting for STDIN, which is bad. :(
+
+        if ($msg->{'type'} =~ m{^message/(?:rfc822|global)\z}i &&
+            defined $msg->{'decoded'} && $msg->{'decoded'} ne '')
+        {
+	  # Ok, so this part is still semi-recursive, since M::SA::Message
+	  # calls M::SA::Message, but we don't subparse the new message,
+	  # and pull a sneaky "steal our child's queue" maneuver to deal
+	  # with it on our own time.  Reference the decoded array directly
+	  # since it's faster.
 	  # 
           my $msg_obj = Mail::SpamAssassin::Message->new({
-    	    message	=>	$toparse->[0]->{'decoded'},
+    	    message	=>	$msg->{'decoded'},
 	    parsenow	=>	0,
 	    normalize	=>	$self->{normalize},
-	    subparse	=>	$toparse->[3]-1,
+	    subparse	=>	$subparse - 1,
 	    });
 
 	  # Add the new message to the current node
-          $toparse->[0]->add_body_part($msg_obj);
+          $msg->add_body_part($msg_obj);
 
 	  # now this is the sneaky bit ... steal the sub-message's parse_queue
 	  # and add it to ours.  then we'll handle the sub-message in our
@@ -670,14 +797,15 @@ sub parse_body {
 	  # Ok, we've subparsed, so go ahead and remove the raw and decoded
 	  # data because we won't need them anymore (the tree under this part
 	  # will have that data)
-	  if (ref $toparse->[0]->{'raw'} eq 'GLOB') {
+	  if (ref $msg->{'raw'} eq 'GLOB') {
 	    # Make sure we close it if it's a temp file -- Bug 5166
-	    close ($toparse->[0]->{'raw'});
+	    close($msg->{'raw'})
+	      or die "error closing input file: $!";
 	  }
 
-	  delete $toparse->[0]->{'raw'};
+	  delete $msg->{'raw'};
 	  
-	  delete $toparse->[0]->{'decoded'};
+	  delete $msg->{'decoded'};
         }
       }
     }
@@ -718,10 +846,20 @@ sub _parse_multipart {
     my $line;
     my $tmp_line = @{$body};
     for ($line=0; $line < $tmp_line; $line++) {
+#     dbg("message: multipart line $line: \"" . $body->[$line] . "\"");
       # specifically look for an opening boundary
-      if ($body->[$line] =~ /^--\Q$boundary\E\s*$/) {
+      if (substr($body->[$line],0,2) eq '--'  # triage
+          && $body->[$line] =~ /^--\Q$boundary\E\s*$/) {
 	# Make note that we found the opening boundary
 	$self->{mime_boundary_state}->{$boundary} = 1;
+
+	# if the line after the opening boundary isn't a header, flag it.
+	# we need to make sure that there's actually another line though.
+	# no re "strict";  # since perl 5.21.8: Ranges of ASCII printables...
+	if ($line+1 < $tmp_line && $body->[$line+1] !~ /^[\041-\071\073-\176]+:/) {
+	  $self->{'missing_mime_headers'} = 1;
+	}
+
         last;
       }
     }
@@ -739,14 +877,32 @@ sub _parse_multipart {
   my $in_body = 0;
   my $header;
   my $part_array;
+  my $found_end_boundary;
 
   my $line_count = @{$body};
   foreach ( @{$body} ) {
     # if we're on the last body line, or we find any boundary marker,
-    # deal with the mime part
-    if ( --$line_count == 0 || (defined $boundary && /^--\Q$boundary\E(?:--)?\s*$/) ) {
+    # deal with the mime part;
+    # a triage before an unlikely-to-match regexp avoids a CPU hotspot
+    $found_end_boundary = defined $boundary && substr($_,0,2) eq '--'
+                          && /^--\Q$boundary\E(?:--)?\s*$/;
+    if ( --$line_count == 0 || $found_end_boundary ) {
       my $line = $_; # remember the last line
 
+      # If at last line and no end boundary found, the line belongs to body
+      # TODO:
+      #  Is $self->{mime_boundary_state}->{$boundary}-- needed here?
+      #  Could "missing end boundary" be a useful rule? Mark it somewhere?
+      #  If SA processed truncated message from amavis etc, this could also
+      #  be hit legimately..
+      if (!$found_end_boundary) {
+        # TODO: This is duplicate code from few pages down below..
+        while (length ($_) > MAX_BODY_LINE_LENGTH) {
+          push (@{$part_array}, substr($_, 0, MAX_BODY_LINE_LENGTH)."\n");
+          substr($_, 0, MAX_BODY_LINE_LENGTH) = '';
+        }
+        push ( @{$part_array}, $_ );
+      }
       # per rfc 1521, the CRLF before the boundary is part of the boundary:
       # NOTE: The CRLF preceding the encapsulation line is conceptually
       # attached to the boundary so that it is possible to have a part
@@ -755,7 +911,7 @@ sub _parse_multipart {
       # CRLFs preceding the encapsulation line, the first of which is part
       # of the preceding body part, and the second of which is part of the
       # encapsulation boundary.
-      if ($part_array) {
+      elsif ($part_array) {
         chomp( $part_array->[-1] );  # trim the CRLF that's part of the boundary
         splice @{$part_array}, -1 if ( $part_array->[-1] eq '' ); # blank line for the boundary only ...
       }
@@ -779,10 +935,19 @@ sub _parse_multipart {
       # rfc 1521 says /^--boundary--$/, some MUAs may just require /^--boundary--/
       # but this causes problems with horizontal lines when the boundary is
       # made up of dashes as well, etc.
-      if (defined $boundary && $line =~ /^--\Q${boundary}\E--\s*$/) {
-	# Make a note that we've seen the end boundary
-	$self->{mime_boundary_state}->{$boundary}--;
-        last;
+      if (defined $boundary) {
+        # no re "strict";  # since perl 5.21.8: Ranges of ASCII printables...
+        if ($line =~ /^--\Q${boundary}\E--\s*$/) {
+	  # Make a note that we've seen the end boundary
+	  $self->{mime_boundary_state}->{$boundary}--;
+          last;
+        }
+	elsif ($line_count && $body->[-$line_count] !~ /^[\041-\071\073-\176]+:/) {
+          # if we aren't on an end boundary and there are still lines left, it
+	  # means we hit a new start boundary.  therefore, the next line ought
+	  # to be a mime header.  if it's not, mark it.
+	  $self->{'missing_mime_headers'} = 1;
+	}
       }
 
       # make sure we start with a new clean node
@@ -796,7 +961,8 @@ sub _parse_multipart {
 
     if (!$in_body) {
       # s/\s+$//;   # bug 5127: don't clean this up (yet)
-      if (m/^[\041-\071\073-\176]+:/) {
+      # no re "strict";  # since perl 5.21.8: Ranges of ASCII printables...
+      if (/^[\041-\071\073-\176]+[ \t]*:/) {
         if ($header) {
           my ( $key, $value ) = split ( /:\s*/, $header, 2 );
           $part_msg->header( $key, $value );
@@ -804,7 +970,7 @@ sub _parse_multipart {
         $header = $_;
 	next;
       }
-      elsif (/^[ \t]/) {
+      elsif (/^[ \t]/ && $header) {
         # $_ =~ s/^\s*//;   # bug 5127, again
         $header .= $_;
 	next;
@@ -828,7 +994,7 @@ sub _parse_multipart {
     }
 
     # we run into a perl bug if the lines are astronomically long (probably
-    # due to lots of regexp backtracking); so cut short any individual line
+    # due to lots of regexp backtracking); so split any individual line
     # over MAX_BODY_LINE_LENGTH bytes in length.  This can wreck HTML
     # totally -- but IMHO the only reason a luser would use
     # MAX_BODY_LINE_LENGTH-byte lines is to crash filters, anyway.
@@ -839,6 +1005,19 @@ sub _parse_multipart {
     push ( @{$part_array}, $_ );
   }
 
+  # Look for a message epilogue
+  # originally ignored whitespace:   0.185   0.2037   0.0654    0.757   0.00   0.00  TVD_TAB
+  # ham FPs were all "." on a line by itself.
+  # spams seem to only have NULL chars afterwards ?
+  if ($line_count) {
+    for(; $line_count > 0; $line_count--) {
+      if ($body->[-$line_count] =~ /[^\s.]/) {
+        $self->{mime_epilogue_exists} = 1;
+        last;
+      }
+    }
+  }
+
 }
 
 =item _parse_normal()
@@ -870,6 +1049,9 @@ sub _parse_normal {
   elsif ($ct[3]) {
     $msg->{'name'} = $ct[3];
   }
+  if ($msg->{'name'}) {
+    $msg->{'name'} = Encode::decode("MIME-Header", $msg->{'name'});
+  }
 
   $msg->{'boundary'} = $boundary;
 
@@ -878,107 +1060,78 @@ sub _parse_normal {
   # up RAM with something we're not going to use?
   #
   if ($msg->{'type'} !~ m@^(?:text/(?:plain|html)$|message\b)@) {
-    my $filepath;
-    ($filepath, $msg->{'raw'}) = Mail::SpamAssassin::Util::secure_tmpfile();
-
-    if ($filepath) {
+    my($filepath, $fh);
+    eval {
+      ($filepath, $fh) = Mail::SpamAssassin::Util::secure_tmpfile();  1;
+    } or do {
+      my $eval_stat = $@ ne '' ? $@ : "errno=$!";  chomp $eval_stat;
+      info("message: failed to create a temp file: %s", $eval_stat);
+    };
+    if ($fh) {
       # The temp file was created, add it to the list of pending deletions
       # we cannot just delete immediately in the POSIX idiom, as this is
       # unportable (to win32 at least)
       push @{$self->{tmpfiles}}, $filepath;
-      $msg->{'raw'}->print(@{$body});
+      dbg("message: storing a message part to file %s", $filepath);
+      $fh->print(@{$body})  or die "error writing to $filepath: $!";
+      $fh->flush  or die "error writing (flush) to $filepath: $!";
+      $msg->{'raw'} = $fh;
     }
   }
 
   # if the part didn't get a temp file, go ahead and store the data in memory
-  if (!exists $msg->{'raw'}) {
+  if (!defined $msg->{'raw'}) {
+    dbg("message: storing a body to memory");
     $msg->{'raw'} = $body;
   }
 }
 
 # ---------------------------------------------------------------------------
 
-sub get_rendered_body_text_array {
+sub get_mimepart_digests {
   my ($self) = @_;
 
-  if (exists $self->{text_rendered}) { return $self->{text_rendered}; }
-
-  $self->{text_rendered} = [];
-
-  # Find all parts which are leaves
-  my @parts = $self->find_parts(qr/./,1);
-  return $self->{text_rendered} unless @parts;
-
-  # the html metadata may have already been set, so let's not bother if it's
-  # already been done.
-  my $html_needs_setting = !exists $self->{metadata}->{html};
-
-  # Go through each part
-  my $text = $self->get_header ('subject') || "\n";
-  for(my $pt = 0 ; $pt <= $#parts ; $pt++ ) {
-    my $p = $parts[$pt];
-
-    # put a blank line between parts ...
-    $text .= "\n";
-
-    my($type, $rnd) = $p->rendered(); # decode this part
-    if ( defined $rnd ) {
-      # Only text/* types are rendered ...
-      $text .= $rnd;
-
-      # TVD - if there are multiple parts, what should we do?
-      # right now, just use the last one.  we may need to give some priority
-      # at some point, ie: use text/html rendered if it exists, or
-      # text/plain rendered as html otherwise.
-      if ($html_needs_setting && $type eq 'text/html') {
-        $self->{metadata}->{html} = $p->{html_results};
-      }
-    }
+  if (!exists $self->{mimepart_digests}) {
+    # traverse all parts which are leaves, recursively
+    $self->{mimepart_digests} =
+      [ map(sha1_hex($_->decode) . ':' . lc($_->{type}||''),
+            $self->find_parts(qr/^/,1,1)) ];
   }
-
-  # whitespace handling (warning: small changes have large effects!)
-  $text =~ s/\n+\s*\n+/\f/gs;		# double newlines => form feed
-  $text =~ tr/ \t\n\r\x0b\xa0/ /s;	# whitespace => space
-  $text =~ tr/\f/\n/;			# form feeds => newline
-  
-  # warn "message: $text";
-
-  my @textary = split_into_array_of_short_lines ($text);
-  $self->{text_rendered} = \@textary;
-
-  return $self->{text_rendered};
+  return $self->{mimepart_digests};
 }
 
 # ---------------------------------------------------------------------------
 
-# TODO: possibly this should just replace get_rendered_body_text_array().
-# (although watch out, this one doesn't copy {html} to metadata)
-sub get_visible_rendered_body_text_array {
-  my ($self) = @_;
+# common code for get_rendered_body_text_array,
+# get_visible_rendered_body_text_array, get_invisible_rendered_body_text_array
+#
+sub get_body_text_array_common {
+  my ($self, $method_name) = @_;
 
-  if (exists $self->{text_visible_rendered}) {
-    return $self->{text_visible_rendered};
-  }
+  my $key = 'text_' . $method_name;
+  if (exists $self->{$key}) { return $self->{$key} }
 
-  $self->{text_visible_rendered} = [];
+  $self->{$key} = [];
 
   # Find all parts which are leaves
   my @parts = $self->find_parts(qr/./,1);
-  return $self->{text_visible_rendered} unless @parts;
+  return $self->{$key} unless @parts;
 
   # the html metadata may have already been set, so let's not bother if it's
   # already been done.
   my $html_needs_setting = !exists $self->{metadata}->{html};
 
+  my $text = $method_name eq 'invisible_rendered' ? ''
+               : ($self->get_header('subject') || "\n");
+
   # Go through each part
-  my $text = $self->get_header ('subject') || "\n";
-  for(my $pt = 0 ; $pt <= $#parts ; $pt++ ) {
+  for (my $pt = 0 ; $pt <= $#parts ; $pt++ ) {
     my $p = $parts[$pt];
 
     # put a blank line between parts ...
-    $text .= "\n";
+    $text .= "\n"  if $text ne '';
 
-    my($type, $rnd) = $p->visible_rendered(); # decode this part
+    my($type, $rnd) = $p->$method_name();  # decode this part
     if ( defined $rnd ) {
       # Only text/* types are rendered ...
       $text .= $rnd;
@@ -995,64 +1148,31 @@ sub get_visible_rendered_body_text_array {
 
   # whitespace handling (warning: small changes have large effects!)
   $text =~ s/\n+\s*\n+/\f/gs;		# double newlines => form feed
-  $text =~ tr/ \t\n\r\x0b\xa0/ /s;	# whitespace => space
+# $text =~ tr/ \t\n\r\x0b\xa0/ /s;	# whitespace (incl. VT, NBSP) => space
+  $text =~ tr/ \t\n\r\x0b/ /s;		# whitespace (incl. VT) => space
   $text =~ tr/\f/\n/;			# form feeds => newline
 
-  my @textary = split_into_array_of_short_lines ($text);
-  $self->{text_visible_rendered} = \@textary;
+  my @textary = split_into_array_of_short_lines($text);
+  $self->{$key} = \@textary;
 
-  return $self->{text_visible_rendered};
+  return $self->{$key};
+}
+
+# ---------------------------------------------------------------------------
+
+sub get_rendered_body_text_array {
+  my ($self) = @_;
+  return $self->get_body_text_array_common('rendered');
+}
+
+sub get_visible_rendered_body_text_array {
+  my ($self) = @_;
+  return $self->get_body_text_array_common('visible_rendered');
 }
 
 sub get_invisible_rendered_body_text_array {
   my ($self) = @_;
-
-  if (exists $self->{text_invisible_rendered}) {
-    return $self->{text_invisible_rendered};
-  }
-
-  $self->{text_invisible_rendered} = [];
-
-  # Find all parts which are leaves
-  my @parts = $self->find_parts(qr/./,1);
-  return $self->{text_invisible_rendered} unless @parts;
-
-  # the html metadata may have already been set, so let's not bother if it's
-  # already been done.
-  my $html_needs_setting = !exists $self->{metadata}->{html};
-
-  # Go through each part
-  my $text = '';
-  for(my $pt = 0 ; $pt <= $#parts ; $pt++ ) {
-    my $p = $parts[$pt];
-
-    # put a blank line between parts ...
-    $text .= "\n" if ( $text );
-
-    my($type, $rnd) = $p->invisible_rendered(); # decode this part
-    if ( defined $rnd ) {
-      # Only text/* types are rendered ...
-      $text .= $rnd;
-
-      # TVD - if there are multiple parts, what should we do?
-      # right now, just use the last one.  we may need to give some priority
-      # at some point, ie: use text/html rendered if it exists, or
-      # text/plain rendered as html otherwise.
-      if ($html_needs_setting && $type eq 'text/html') {
-        $self->{metadata}->{html} = $p->{html_results};
-      }
-    }
-  }
-
-  # whitespace handling (warning: small changes have large effects!)
-  $text =~ s/\n+\s*\n+/\f/gs;		# double newlines => form feed
-  $text =~ tr/ \t\n\r\x0b\xa0/ /s;	# whitespace => space
-  $text =~ tr/\f/\n/;			# form feeds => newline
-
-  my @textary = split_into_array_of_short_lines ($text);
-  $self->{text_invisible_rendered} = \@textary;
-
-  return $self->{text_invisible_rendered};
+  return $self->get_body_text_array_common('invisible_rendered');
 }
 
 # ---------------------------------------------------------------------------
@@ -1074,7 +1194,8 @@ sub get_decoded_body_text_array {
     next if ($parts[$pt]->{'type'} eq 'text/calendar');
 
     push(@{$self->{text_decoded}}, "\n") if ( @{$self->{text_decoded}} );
-    push(@{$self->{text_decoded}}, $parts[$pt]->decode());
+    push(@{$self->{text_decoded}},
+         split_into_array_of_short_paragraphs($parts[$pt]->decode()));
   }
 
   return $self->{text_decoded};
@@ -1083,10 +1204,10 @@ sub get_decoded_body_text_array {
 # ---------------------------------------------------------------------------
 
 sub split_into_array_of_short_lines {
-  my @result = ();
+  my @result;
   foreach my $line (split (/^/m, $_[0])) {
     while (length ($line) > MAX_BODY_LINE_LENGTH) {
-      # try splitting "nicely" so that we don't chop an url in half or
+      # try splitting "nicely" so that we don't chop a url in half or
       # something.  if there's no space, then just split at max length.
       my $length = rindex($line, ' ', MAX_BODY_LINE_LENGTH) + 1;
       $length ||= MAX_BODY_LINE_LENGTH;
@@ -1099,6 +1220,28 @@ sub split_into_array_of_short_lines {
 
 # ---------------------------------------------------------------------------
 
+# split a text into array of paragraphs of sizes between
+# $chunk_size and 2 * $chunk_size, returning the resulting array
+
+sub split_into_array_of_short_paragraphs {
+  my @result;
+  my $chunk_size = 1024;
+  my $text_l = length($_[0]);
+  my($j,$ofs);
+  for ($ofs = 0;  $text_l - $ofs > 2 * $chunk_size;  $ofs = $j+1) {
+    $j = index($_[0], "\n", $ofs+$chunk_size);
+    if ($j < 0) {
+      $j = index($_[0], " ", $ofs+$chunk_size);
+      if ($j < 0) { $j = $ofs+$chunk_size }
+    }
+    push(@result, substr($_[0], $ofs, $j-$ofs+1));
+  }
+  push(@result, substr($_[0], $ofs))  if $ofs < $text_l;
+  @result;
+}
+
+# ---------------------------------------------------------------------------
+
 1;
 
 =back
