@@ -45,6 +45,8 @@ use warnings;
 use bytes;
 use re 'taint';
 
+require 5.008001;  # needs utf8::is_utf8()
+
 use Mail::SpamAssassin::Logger;
 
 BEGIN {
@@ -58,11 +60,12 @@ BEGIN {
   @ISA = qw(Exporter);
   @EXPORT = ();
   @EXPORT_OK = qw(&local_tz &base64_decode &untaint_var &untaint_file_path
-                  &exit_status_str &proc_status_ok &am_running_on_windows);
+                  &exit_status_str &proc_status_ok &am_running_on_windows
+                  &reverse_ip_address &decode_dns_question_entry
+                  &secure_tmpfile &secure_tmpdir &uri_list_canonicalize);
 }
 
 use Mail::SpamAssassin;
-use Mail::SpamAssassin::Util::RegistrarBoundaries;
 
 use Config;
 use IO::Handle;
@@ -70,6 +73,7 @@ use File::Spec;
 use File::Basename;
 use Time::Local;
 use Sys::Hostname (); # don't import hostname() into this namespace!
+use NetAddr::IP 4.000;
 use Fcntl;
 use Errno qw(ENOENT EACCES EEXIST);
 use POSIX qw(:sys_wait_h WIFEXITED WIFSIGNALED WIFSTOPPED WEXITSTATUS
@@ -80,6 +84,16 @@ use POSIX qw(:sys_wait_h WIFEXITED WIFSIGNALED WIFSTOPPED WEXITSTATUS
 use constant HAS_MIME_BASE64 => eval { require MIME::Base64; };
 use constant RUNNING_ON_WINDOWS => ($^O =~ /^(?:mswin|dos|os2)/oi);
 
+# These are not implemented on windows (see bug 6798 and 6470)
+BEGIN {
+  if (RUNNING_ON_WINDOWS) {
+    *WIFEXITED   = sub { not $_[0] & 127 };
+    *WEXITSTATUS = sub { $_[0] >> 8 };
+    *WIFSIGNALED = sub { ($_[0] & 127) && ($_[0] & 127 != 127) };
+    *WTERMSIG    = sub { $_[0] & 127 };
+  }
+}
+
 ###########################################################################
 
 # find an executable in the current $PATH (or whatever for that platform)
@@ -94,7 +108,7 @@ use constant RUNNING_ON_WINDOWS => ($^O =~ /^(?:mswin|dos|os2)/oi);
     if ( !$displayed_path++ ) {
       dbg("util: current PATH is: ".join($Config{'path_sep'},File::Spec->path()));
     }
-    foreach my $path (File::Spec->path()) {
+    foreach my $path (File::Spec->path(), qw(/usr/local/bin /usr/local/sbin)) {
       my $fname = File::Spec->catfile ($path, $filename);
       if ( -f $fname ) {
         if (-x $fname) {
@@ -106,7 +120,7 @@ use constant RUNNING_ON_WINDOWS => ($^O =~ /^(?:mswin|dos|os2)/oi);
         }
       }
     }
-    return undef;
+    return;
   }
 }
 
@@ -225,11 +239,12 @@ sub untaint_file_path {
   local ($1);
   # Barry Jaspan: allow ~ and spaces, good for Windows.  Also return ''
   # if input is '', as it is a safe path.
-  my $chars = '-_A-Za-z\xA0-\xFF0-9\.\%\@\=\+\,\/\\\:';
+  my $chars = '-_A-Za-z0-9\xA0-\xFF\.\%\@\=\+\,\/\\\:';
   my $re = qr/^\s*([$chars][${chars}~ ]*)$/o;
 
   if ($path =~ $re) {
-    return untaint_var($1);
+    $path = $1;
+    return untaint_var($path);
   } else {
     warn "util: refusing to untaint suspicious path: \"$path\"\n";
     return $path;
@@ -268,36 +283,46 @@ sub untaint_hostname {
 #  untaint_var(\%ENV);
 #
 sub untaint_var {
-  no re 'taint';  # override a  "use re 'taint'"  from outer scope
-  local ($_) = @_;
-  return undef unless defined;
-
-  unless (ref) {
+# my $arg = $_[0];  # avoid copying unnecessarily
+  if (!ref $_[0]) { # optimized by-far-the-most-common case
+    no re 'taint';  # override a  "use re 'taint'"  from outer scope
+    return undef if !defined $_[0]; ## no critic (ProhibitExplicitReturnUndef)  - See Bug 7120
     local($1); # avoid Perl taint bug: tainted global $1 propagates taintedness
-    /^(.*)\z/s;
+    $_[0] =~ /^(.*)\z/s;
     return $1;
-  }
-  elsif (ref eq 'ARRAY') {
-    $_ = untaint_var($_)  for @{$_};
-    return @{$_} if wantarray;
-  }
-  elsif (ref eq 'HASH') {
-    while (my ($k, $v) = each %{$_}) {
-      if (!defined $v && $_ == \%ENV) {
-	delete ${$_}{$k};
-	next;
+  } else {
+    my $r = ref $_[0];
+    if ($r eq 'ARRAY') {
+      my $arg = $_[0];
+      $_ = untaint_var($_)  for @{$arg};
+      return @{$arg} if wantarray;
+    }
+    elsif ($r eq 'HASH') {
+      my $arg = $_[0];
+      if ($arg == \%ENV) {  # purge undefs from %ENV, untaint the rest
+        while (my($k, $v) = each %{$arg}) {
+          # It is safe to delete the item most recently returned by each()
+          if (!defined $v) { delete ${$arg}{$k}; next }
+          ${$arg}{untaint_var($k)} = untaint_var($v);
+        }
+      } else {
+        # hash keys are never tainted,
+        # although old version of perl had some quirks there
+        while (my($k, $v) = each %{$arg}) {
+          ${$arg}{untaint_var($k)} = untaint_var($v);
+        }
       }
-      ${$_}{untaint_var($k)} = untaint_var($v);
+      return %{$arg} if wantarray;
+    }
+    elsif ($r eq 'SCALAR' || $r eq 'REF') {
+      my $arg = $_[0];
+      ${$arg} = untaint_var(${$arg});
+    }
+    else {
+      warn "util: can't untaint a $r !\n";
     }
-    return %{$_} if wantarray;
   }
-  elsif (ref eq 'SCALAR' or ref eq 'REF') {
-    ${$_} = untaint_var(${$_});
-  }
-  else {
-    warn "util: can't untaint a " . ref($_) . "!\n";
-  }
-  return $_;
+  return $_[0];
 }
 
 ###########################################################################
@@ -317,13 +342,11 @@ sub taint_var {
 # append optional mesage (dual-valued errno or a string or a number),
 # returning the resulting string
 #
-sub exit_status_str($;$) {
+sub exit_status_str {
   my($stat,$errno) = @_;
   my $str;
   if (!defined($stat)) {
     $str = '(no status)';
-  } elsif (am_running_on_windows()) { 
-    $str = 'exit (running under Windows, cannot determine exit status)'
   } elsif (WIFEXITED($stat)) {
     $str = sprintf("exit %d", WEXITSTATUS($stat));
   } elsif (WIFSTOPPED($stat)) {
@@ -347,7 +370,7 @@ sub exit_status_str($;$) {
 # check errno to be 0 and a process exit status to be in the list of success
 # status codes, returning true if both are ok, and false otherwise
 #
-sub proc_status_ok($;$@) {
+sub proc_status_ok {
   my($exit_status,$errno,@success) = @_;
   my $ok = 0;
   if ((!defined $errno || $errno == 0) && WIFEXITED($exit_status)) {
@@ -457,7 +480,7 @@ sub parse_rfc822_date {
     $dd = $1; $mon = lc($2); $yyyy = $3;
   } else {
     dbg("util: time cannot be parsed: $date");
-    return undef;
+    return;
   }
 
   # handle two and three digit dates as specified by RFC 2822
@@ -545,7 +568,7 @@ sub parse_rfc822_date {
   } or do {
     my $eval_stat = $@ ne '' ? $@ : "errno=$!";  chomp $eval_stat;
     dbg("util: time cannot be parsed: $date, $yyyy-$mmm-$dd $hh:$mm:$ss, $eval_stat");
-    return undef;
+    return;
   };
 
   if ($tzoff =~ /([-+])(\d\d)(\d\d)$/)	# convert to seconds difference
@@ -706,8 +729,16 @@ sub base64_decode {
 sub qp_decode {
   local $_ = shift;
 
-  s/\=\r?\n//gs;
-  s/\=([0-9a-fA-F]{2})/chr(hex($1))/ge;
+  # RFC 2045: when decoding a Quoted-Printable body, any trailing
+  # white space on a line must be deleted
+  s/[ \t]+(?=\r?\n)//gs;
+
+  s/=\r?\n//gs;  # soft line breaks
+
+  # RFC 2045 explicitly prohibits lowercase characters a-f in QP encoding
+  # do we really want to allow them???
+  s/=([0-9a-fA-F]{2})/chr(hex($1))/ge;
+
   return $_;
 }
 
@@ -783,9 +814,8 @@ sub extract_ipv4_addr_from_string {
     if (defined $1) { return $1; }
   }
 
-  # ignore native IPv6 addresses; currently we have no way to deal with
-  # these if we could extract them, as the DNSBLs don't provide a way
-  # to query them!  TODO, eventually, once IPv6 spam starts to appear ;)
+  # ignore native IPv6 addresses;
+  # TODO, eventually, once IPv6 spam starts to appear ;)
   return;
 }
 
@@ -863,10 +893,61 @@ sub ips_match_in_24_mask {
 
 ###########################################################################
 
+# Given a quad-dotted IPv4 address or an IPv6 address, reverses the order
+# of its bytes (IPv4) or nibbles (IPv6), joins them with dots, producing
+# a string suitable for reverse DNS lookups. Returns undef in case of a
+# syntactically invalid IP address.
+#
+sub reverse_ip_address {
+  my ($ip) = @_;
+
+  my $revip;
+  local($1,$2,$3,$4);
+  if ($ip =~ /^(\d{1,3})\.(\d{1,3})\.(\d{1,3})\.(\d{1,3})\z/) {
+    $revip = "$4.$3.$2.$1";
+  } elsif ($ip !~ /:/ || $ip !~ /^[0-9a-fA-F:.]{2,}\z/) {  # triage
+    # obviously unrecognized syntax
+  } elsif (!NetAddr::IP->can('full6')) {  # since NetAddr::IP 4.010
+    info("util: version of NetAddr::IP is too old, IPv6 not supported");
+  } else {
+    # looks like an IPv6 address, let NetAddr::IP check the details
+    my $ip_obj = NetAddr::IP->new6($ip);
+    if (defined $ip_obj) {  # valid IPv6 address
+      # RFC 5782 section 2.4.
+      $revip = lc $ip_obj->network->full6;  # string in a canonical form
+      $revip =~ s/://g;
+      $revip = join('.', reverse split(//,$revip));
+    }
+  }
+  return $revip;
+}
+
+###########################################################################
+
 sub my_inet_aton { unpack("N", pack("C4", split(/\./, $_[0]))) }
 
 ###########################################################################
 
+sub decode_dns_question_entry {
+  # decodes a Net::DNS::Packet->question entry,
+  # returning a triple: class, type, label
+  #
+  my $q = $_[0];
+  my $qname = $q->qname;
+
+  # Bug 6959, Net::DNS flags a domain name in a query section as utf8, while
+  # still keeping it "RFC 1035 zone file format"-encoded, silly and harmful
+  utf8::encode($qname) if utf8::is_utf8($qname);  # since Perl 5.8.1
+
+  local $1;
+  # Net::DNS provides a query in encoded RFC 1035 zone file format, decode it!
+  $qname =~ s{ \\ ( [0-9]{3} | [^0-9] ) }
+             { length($1)==1 ? $1 : $1 <= 255 ? chr($1) : "\\$1" }xgse;
+  return ($q->qclass, $q->qtype, $qname);
+}
+
+###########################################################################
+
 sub parse_content_type {
   # This routine is typically called by passing a
   # get_header("content-type") which passes all content-type headers
@@ -930,7 +1011,7 @@ sub parse_content_type {
   }
 
   # strip inappropriate chars (bug 5399: after the text/plain fixup)
-  $ct =~ tr/\000-\040\177-\377\042\050\051\054\056\072-\077\100\133-\135//d;
+  $ct =~ tr/\000-\040\177-\377\042\050\051\054\072-\077\100\133-\135//d;
 
   # Now that the header has been parsed, return the requested information.
   # In scalar context, just the MIME type, in array context the
@@ -961,6 +1042,7 @@ sub url_encode {
     }
     # other stuff
     else {
+      # no re "strict";  # since perl 5.21.8
       # 0x00-0x20, 0x7f-0xff, ", %, <, >
       s/([\000-\040\177-\377\042\045\074\076])
 	  /push(@encoded, $1) && sprintf "%%%02x", unpack("C",$1)/egx;
@@ -1017,17 +1099,13 @@ If it cannot open a file after 20 tries, it returns C<undef>.
 sub secure_tmpfile {
   my $tmpdir = untaint_file_path($ENV{'TMPDIR'} || File::Spec->tmpdir());
 
-  if (!$tmpdir) {
-    # Note: we would prefer to keep this fatal, as not being able to
-    # find a writable tmpdir is a big deal for the calling code too.
-    # That would be quite a psychotic case, also.
-    warn "util: cannot find a temporary directory, set TMP or TMPDIR in environment";
-    return;
-  }
+  defined $tmpdir && $tmpdir ne ''
+    or die "util: cannot find a temporary directory, set TMP or TMPDIR in environment";
 
-  my ($reportfile, $tmpfile);
-  my $umask = umask 077;
+  opendir(my $dh, $tmpdir) or die "Could not open directory $tmpdir: $!";
+  closedir $dh or die "Error closing directory $tmpdir: $!";
 
+  my ($reportfile, $tmpfh);
   for (my $retries = 20; $retries > 0; $retries--) {
     # we do not rely on the obscurity of this name for security,
     # we use a average-quality PRG since this is all we need
@@ -1037,33 +1115,35 @@ sub secure_tmpfile {
 
     # instead, we require O_EXCL|O_CREAT to guarantee us proper
     # ownership of our file, read the open(2) man page
-    if (sysopen($tmpfile, $reportfile, O_RDWR|O_CREAT|O_EXCL, 0600)) {
-      binmode $tmpfile  or die "cannot set $reportfile to binmode: $!";
+    if (sysopen($tmpfh, $reportfile, O_RDWR|O_CREAT|O_EXCL, 0600)) {
+      binmode $tmpfh  or die "cannot set $reportfile to binmode: $!";
       last;
     }
-
-    if ($!{EEXIST}) {
-      # it is acceptable if $tmpfile already exists, try another
-      next;
-    }
-    
-    # error, maybe "out of quota" or "too many open files" (bug 4017)
-    warn "util: secure_tmpfile failed to create file '$reportfile': $!\n";
+    my $errno = $!;
 
     # ensure the file handle is not semi-open in some way
-    if ($tmpfile) {
-      close $tmpfile  or info("error closing $reportfile: $!");
+    if ($tmpfh) {
+      if (! close $tmpfh) {
+       info("error closing $reportfile: $!");
+       undef $tmpfh;
+      }
     }
+
+    # it is acceptable if $tmpfh already exists, try another
+    next if $errno == EEXIST;
+
+    # error, maybe "out of quota", "too many open files", "Permission denied"
+    # (bug 4017); makes no sense retrying
+    die "util: failed to create a temporary file '$reportfile': $errno";
   }
 
-  umask $umask;
-
-  if (!$tmpfile) {
-    warn "util: secure_tmpfile failed to create file, giving up";
-    return;	# undef
+  if (!$tmpfh) {
+    warn "util: secure_tmpfile failed to create a temporary file, giving up";
+    return;
   }
 
-  return ($reportfile, $tmpfile);
+  dbg("util: secure_tmpfile created a temporary file %s", $reportfile);
+  return ($reportfile, $tmpfh);
 }
 
 =item my ($dirpath) = secure_tmpdir();
@@ -1123,27 +1203,33 @@ sub secure_tmpdir {
 
 ###########################################################################
 
+##
+## DEPRECATED FUNCTION, only left for third party plugins as fallback.
+## Replaced with Mail::SpamAssassin::RegistryBoundaries::uri_to_domain.
+##
 sub uri_to_domain {
   my ($uri) = @_;
 
   # Javascript is not going to help us, so return.
   return if ($uri =~ /^javascript:/i);
 
-  $uri =~ s,#.*$,,gs;			# drop fragment
-  $uri =~ s#^[a-z]+:/{0,2}##gsi;	# drop the protocol
-  $uri =~ s,^[^/]*\@,,gs;		# username/passwd
+  $uri =~ s{\#.*$}{}gs;			# drop fragment
+  $uri =~ s{^[a-z]+:/{0,2}}{}gsi;	# drop the protocol
+  $uri =~ s{^[^/]*\@}{}gs;		# username/passwd
 
   # strip path and CGI params.  note: bug 4213 shows that "&" should
   # *not* be likewise stripped here -- it's permitted in hostnames by
   # some common MUAs!
-  $uri =~ s,[/\?].*$,,gs;              
+  $uri =~ s{[/?].*$}{}gs;              
 
-  $uri =~ s,:\d*$,,gs;			# port, bug 4191: sometimes the # is missing
+  $uri =~ s{:\d*$}{}gs;		# port, bug 4191: sometimes the # is missing
 
   # skip undecoded URIs if the encoded bits shouldn't be.
   # we'll see the decoded version as well.  see url_encode()
   return if $uri =~ /\%(?:2[1-9a-fA-F]|[3-6][0-9a-fA-F]|7[0-9a-eA-E])/;
 
+  my $host = $uri;  # unstripped/full domain name
+
   # keep IPs intact
   if ($uri !~ /^\d+\.\d+\.\d+\.\d+$/) { 
     # get rid of hostname part of domain, understanding delegation
@@ -1154,11 +1240,12 @@ sub uri_to_domain {
         (Mail::SpamAssassin::Util::RegistrarBoundaries::is_domain_valid($uri));
   }
   
-  # $uri is now the domain only
-  return lc $uri;
+  # $uri is now the domain only, optionally return unstripped host name
+  return !wantarray ? lc $uri : (lc $uri, lc $host);
 }
 
-sub uri_list_canonify {
+*uri_list_canonify = \&uri_list_canonicalize;  # compatibility alias
+sub uri_list_canonicalize {
   my($redirector_patterns, @uris) = @_;
 
   # make sure we catch bad encoding tricks
@@ -1184,15 +1271,15 @@ sub uri_list_canonify {
     # bug 4390: certain MUAs treat back slashes as front slashes.
     # since backslashes are supposed to be encoded in a URI, swap non-encoded
     # ones with front slashes.
-    $nuri =~ tr@\\@/@;
+    $nuri =~ tr{\\}{/};
 
     # http:www.foo.biz -> http://www.foo.biz
-    $nuri =~ s#^(https?:)/{0,2}#$1//#i;
+    $nuri =~ s{^(https?:)/{0,2}}{$1//}i;
 
     # *always* make a dup with all %-encoding decoded, since
     # important parts of the URL may be encoded (such as the
     # scheme). (bug 4213)
-    if ($nuri =~ /\%[0-9a-fA-F]{2}/) {
+    if ($nuri =~ /%[0-9a-fA-F]{2}/) {
       $nuri = Mail::SpamAssassin::Util::url_encode($nuri);
     }
 
@@ -1200,15 +1287,15 @@ sub uri_list_canonify {
     # unschemed URIs: assume default of "http://" as most MUAs do
     if ($nuri !~ /^[-_a-z0-9]+:/i) {
       if ($nuri =~ /^ftp\./) {
-	$nuri =~ s@^@ftp://@g;
+	$nuri =~ s{^}{ftp://}g;
       }
       else {
-	$nuri =~ s@^@http://@g;
+	$nuri =~ s{^}{http://}g;
       }
     }
 
     # http://www.foo.biz?id=3 -> http://www.foo.biz/?id=3
-    $nuri =~ s@^(https?://[^/?]+)\?@$1/?@i;
+    $nuri =~ s{^(https?://[^/?]+)\?}{$1/?}i;
 
     # deal with encoding of chars, this is just the set of printable
     # chars minus ' ' (that is, dec 33-126, hex 21-7e)
@@ -1227,6 +1314,22 @@ sub uri_list_canonify {
       # not required
       $rest ||= '';
 
+      # Bug 6751:
+      # RFC 3490 (IDNA): Whenever dots are used as label separators, the
+      #   following characters MUST be recognized as dots: U+002E (full stop),
+      #   U+3002 (ideographic full stop), U+FF0E (fullwidth full stop),
+      #   U+FF61 (halfwidth ideographic full stop).
+      # RFC 5895: [...] the IDEOGRAPHIC FULL STOP character (U+3002)
+      #   can be mapped to the FULL STOP before label separation occurs.
+      #   [...] Only the IDEOGRAPHIC FULL STOP character (U+3002) is added in
+      #   this mapping because the authors have not fully investigated [...]
+      # Adding also 'SMALL FULL STOP' (U+FE52) as seen in the wild.
+      # Parhaps also the 'ONE DOT LEADER' (U+2024).
+      if ($host =~ s{(?: \xE3\x80\x82 | \xEF\xBC\x8E | \xEF\xBD\xA1 |
+                         \xEF\xB9\x92 | \xE2\x80\xA4 )}{.}xgs) {
+        push(@nuris, join ('', $proto, $host, $rest));
+      }
+
       # bug 4146: deal with non-US ASCII 7-bit chars in the host portion
       # of the URI according to RFC 1738 that's invalid, and the tested
       # browsers (Firefox, IE) remove them before usage...
@@ -1332,7 +1435,7 @@ sub first_date {
     my $time = parse_rfc822_date($string);
     return $time if defined($time) && $time;
   }
-  return undef;
+  return;
 }
 
 sub receive_date {
@@ -1580,9 +1683,9 @@ sub regexp_remove_delimiters {
   my $delim;
   if (!defined $re || $re eq '') {
     warn "cannot remove delimiters from null regexp";
-    return undef;   # invalid
+    return;  # invalid
   }
-  elsif ($re =~ s/^m{//) {              # m{foo/bar}
+  elsif ($re =~ s/^m\{//) {             # m{foo/bar}
     $delim = '}';
   }
   elsif ($re =~ s/^m\(//) {             # m(foo/bar)
@@ -1649,7 +1752,7 @@ sub avoid_db_file_locking_bug {
     my $stat_errn = stat($file) ? 0 : 0+$!;
     next if $stat_errn == ENOENT;
 
-    dbg("Berkeley DB bug work-around: cleaning tmp file $file");
+    dbg("util: Berkeley DB bug work-around: cleaning tmp file $file");
     unlink($file) or warn "cannot remove Berkeley DB tmp file $file: $!\n";
   }
 }
